VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CTreeNode"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit

'****************************************************
'* CTreeNode.cls                                    *
'* Standalone TreeNode class                        *
'* Programmed: Konrad Rudolph, Mad Rat Pro.         *
'* Last Change: 20.1.2003 23:27                     *
'* Version: 1.0.0                                   *
'****************************************************

'----------------------------------------------------
' INFORMATION
'----------------------------------------------------
'
' This class can be used to save data in hierarchised
' tree structures. Each tree node has all abilities
' of an own tree, thus it provides an enormous flexibility.
' Examle, if you have a Tree called Tree1, a Node called
' "Node_1" with a sub-node called Node_1.1 and you want to
' give this one a sub-node, called "Node_1.1.1".
' This is the code you would use:
'
'   Tree1.Item("Node_1").Item("Node_1.1").Add "Node_1.1.1", "Value"
'
' additionally, yould make use of the standard property
' so the above statement could become
'
'   Tree1("Node_1")("Node_1.1").Add "Node_1.1.1", "Value"
'
' or, further:
'
'   Tree1!Node_1![Node_1.1].Add "Node_1.1.1", "Value".
'
' The class supports also path finding, this means you can
' enter all keys in one string, with backslash as separator.
' The following statement would to the same as above:
'
'   Tree1.Item("Node_1\Node_1.1").Add "Node_1.1.1", "Value"
'
' notice, however, that the last way is minutely slower due
' to the parsing progress of the key string.
' The above methods (including path finding) work with all
' functions, provided by the CTreeNode class, that require
' a 'key' parametre.
'
'----------------------------------------------------

Private mValue      As String           ' the node's value
Private mKey        As String           ' a level-unique identifier

Private mCount      As Long             ' number of first-level children

Private mNext_      As CTreeNode        ' pointer to next node on same level
Private mPrev_      As CTreeNode        ' pointer to previous node on same level

Private mFirst      As CTreeNode        ' pointer to first child element
Private mLast       As CTreeNode        ' pointer to last child element

Private mParent     As CTreeNode        ' pointer to parent node
Private mChildren   As CTreeNode        ' pointer to children
'----------------------------------------------------

' returns the number of first-level child nodes
Public Property Get Count() As Long
    Count = mCount
End Property

' sets or returns the node's value
Public Property Get Value() As String
    Value = mValue
End Property
Public Property Let Value(Value As String)
    mValue = Value
End Property

'same for key
Public Property Get Key() As String
    Key = mKey
End Property
Friend Property Let Key(Value As String)
    mKey = Value
End Property

' same for parental node
Public Property Get Parent() As CTreeNode
    Set Parent = mParent
End Property
Friend Property Set Parent(Value As CTreeNode)
    Set mParent = Value
End Property

' same for next node on same tree level
Public Property Get Next_() As CTreeNode
    Set Next_ = mNext_
End Property
Friend Property Set Next_(Value As CTreeNode)
    Set mNext_ = Value
End Property

'  same for previous node on same tree level
Public Property Get Prev_() As CTreeNode
    Set Prev_ = mPrev_
End Property
Friend Property Set Prev_(Value As CTreeNode)
    Set mPrev_ = Value
End Property

' same for first node on this tree level
Public Property Get First() As CTreeNode
    Set First = mFirst
End Property
Friend Property Set First(Value As CTreeNode)
    Set mFirst = Value
End Property

' same for last node on this tree level
Public Property Get Last() As CTreeNode
    Set Last = mLast
End Property
Friend Property Set Last(Value As CTreeNode)
    Set mLast = Value
End Property

' returns the tree structure that recursively
' describes all child nodes of this one
Public Property Get Nodes() As CTreeNode
    If Not mChildren Is Nothing Then _
        Set Nodes = mChildren.First
End Property
Friend Property Set Nodes(Value As CTreeNode)
    If Not mChildren Is Nothing Then _
        Set mChildren.First = Value
End Property
'----------------------------------------------------
'----------------------------------------------------
' Add As CTreeNode
' Adds a new node to this one as a child
' - Key As String - provides an identifier for the node
' - Value As String - value of the new node
' returns a reference to the newly created node
'----------------------------------------------------
Public Function Add( _
  Key As String, _
  Optional Value As String _
) As CTreeNode
    Dim Prim    As String, _
        Append  As String
    
    Prim = Key
    If ResolveKey(Prim, Append) Then
        ' proceed sub-key call
        Set Add = Item(Prim).Add(Append, Value)
        Exit Function
    End If
    
    Set Add = New CTreeNode
    
    With Add
        .Value = Value
        .Key = Key
        Set .Parent = Me
    End With
    
    If mChildren Is Nothing Then _
        Set mChildren = New CTreeNode
    
    If mCount Then
        With mChildren
            Set .Last.Next_ = Add
            Set Add.Prev_ = .Last
            Set .Last = Add
        End With
    Else
        ' create first node
        With mChildren
            Set .First = Add
            Set .Last = Add
        End With
    End If
    
    mCount = mCount + 1
End Function

'----------------------------------------------------
' Remove As Boolean
' removes a specified node
' - Key As String - describes the node to remove
'   if more nodes with the same key are present only
'   the first one found is erased
' removes success
'----------------------------------------------------
Public Function Remove(Key As String) As Boolean
    Dim tmpNode As CTreeNode
    Dim Prim    As String, _
        Append  As String
    
    Prim = Key
    If ResolveKey(Prim, Append) Then
        ' proceed sub-key call
        Remove = Item(Prim).Remove(Append)
        Exit Function
    End If
    
    If mCount = 0 Then _
        Exit Function
    
    Set tmpNode = mChildren.First
    
    Do Until tmpNode Is Nothing
        If tmpNode.Key = Key Then
            If Not tmpNode.Next_ Is Nothing Then
                If Not tmpNode.Prev_ Is Nothing Then
                    Set tmpNode.Prev_.Next_ = tmpNode.Next_
                    Set tmpNode.Next_.Prev_ = tmpNode.Prev_
                Else
                    Set mChildren.First = tmpNode.Next_
                    Set tmpNode.Next_.Prev_ = Nothing
                End If
            Else
                If Not tmpNode.Prev_ Is Nothing Then
                    Set tmpNode.Prev_.Next_ = Nothing
                    Set mChildren.Last = tmpNode.Next_
                Else
                    Set mChildren.First = Nothing
                    Set mChildren.Last = Nothing
                End If
            End If
            
            Set tmpNode = Nothing
            Remove = True
            mCount = mCount - 1
        Else
            Set tmpNode = tmpNode.Next_
        End If
    Loop
End Function

'----------------------------------------------------
' Item As CTreeNode
' returns a reference to a specified item
' - Key As String - a specifier
' returns the first child item found with specified key
'----------------------------------------------------
' Procedure-Attribute (Default)
Public Function Item(Key As String) As CTreeNode
Attribute Item.VB_UserMemId = 0
    Dim tmpNode As CTreeNode
    Dim Prim    As String, _
        Append  As String
    
    If mChildren Is Nothing Then _
        Exit Function
    
    Prim = Key
    If ResolveKey(Prim, Append) Then
        ' proceed sub-key call
        Set Item = Item(Prim).Item(Append)
        Exit Function
    End If
    
    Set tmpNode = mChildren.First
    
    Do Until tmpNode Is Nothing
        If tmpNode.Key = Prim Then
            If LenB(Append) Then
                Set Item = tmpNode.Item(Append)
            Else
                Set Item = tmpNode
            End If
            
            Set tmpNode = Nothing
        Else
            Set tmpNode = tmpNode.Next_
        End If
    Loop
End Function

'----------------------------------------------------
' Clear
' erases whole sub-tree
'----------------------------------------------------
Public Sub Clear()
    Set mChildren = New CTreeNode
    mCount = 0
End Sub
'----------------------------------------------------

'----------------------------------------------------
' ExportXML As Boolean
' exports the subtree's content as XML to a file
' - Path As String - path of the xml file to create
' - Key As String - indicates the starting node
' returns success
'----------------------------------------------------
Public Function ExportXML( _
  Path As String, _
  Optional Key As String _
) As Boolean
    Dim Content As String
    
    If LenB(Key) Then
        ' proceed sub-key call
        ExportXML = Item(Key).ExportXML(Path, "")
        Exit Function
    Else
        Content = "<?xml version=""1.0"" ?>" & vbCrLf & vbCrLf
        
        Content = Content & NodesToString(Me)
        
        With New CFile
            If .OpenFile(Path, OF_CREATENEW) Then
                .OutputStream.WriteString Content
                .CloseFile
                
                ExportXML = True
            End If
        End With
    End If
End Function

'----------------------------------------------------
' ImportXML As Boolean
' imports an XML file and fills the subtree with it
' - Path As String - path of the XML file to parse
' returns success
'----------------------------------------------------
Public Function ImportXML( _
  Path As String _
) As Boolean
    With New CFile
        If .OpenFile(Path, OF_OPENEXISTING) Then
            Call Clear
            
            ImportXML = ParseXML(.InputStream.ReadFile, "", "")
            
            .CloseFile
        End If
    End With
End Function

'----------------------------------------------------
'----------------------------------------------------
' private helper functions
'----------------------------------------------------

Private Function ResolveKey( _
  ByRef Key As String, _
  ByRef Append As String _
) As Boolean
    Dim Slash   As Long
    
    Slash = InStrRev(Key, "\")
    If Slash Then
        ' key contains subkeys
        
        ' remove trailing backslashes
        If RightB$(Key, 2) = "\" Then _
            Key = LeftB$(Key, LenB(Key) - 2)
        
        ' split key
        Append = MidB$(Key, Slash * 2 + 1, LenB(Key) - Slash * 2)
        Key = MidB$(Key, 1, Slash * 2 - 2)
        
        ResolveKey = True
    End If
End Function

Private Function NodesToString(Start As CTreeNode) As String
    Dim Node        As CTreeNode
    Dim NTS         As String
    Static Level    As Long
    
    Set Node = Start
    
    Do Until Node Is Nothing
        With Node
            If .Count Then
                NTS = NTS & String$(Level * 4, " ") & _
                  "<" & .Key & ">" & vbCrLf & _
                  IIf(LenB(.Value), _
                  String$(Level * 4 + 4, " ") & ReplaceEntities(.Value) & vbCrLf, "")
                
                Level = Level + 1
                NTS = NTS & NodesToString(.Nodes)
                Level = Level - 1
                
                NTS = NTS & String$(Level * 4, " ") & "</" & .Key & ">" & vbCrLf
            Else
                NTS = NTS & String$(Level * 4, " ") & _
                  "<" & .Key & ">" & ReplaceEntities(.Value) & _
                  "</" & .Key & ">" & vbCrLf
            End If
            
            Set Node = .Next_
        End With
        
        If Level = 0 Then _
            Exit Do
    Loop
    
    NodesToString = NTS
End Function

Private Function ParseXML( _
  Content As String, _
  CurrentKey As String, _
  RootPath As String _
) As Boolean
    Dim I           As Long
    Dim Char        As String * 1, _
        LastChar    As String * 1
    Dim Buff        As String, _
        CurrKey     As String, _
        CurrValue   As String, _
        Path        As String
    Dim InBraket    As Boolean, _
        BracketEnd  As Boolean, _
        EmptyEnd    As Boolean, _
        InRemark    As Boolean, _
        InKey       As Boolean, _
        InString    As Boolean
    
    CurrKey = CurrentKey
    Path = RootPath
    
    For I = 1 To LenB(Content) Step 2
        Char = MidB$(Content, I, 2)
        
        If InRemark Then
            Select Case Char
            Case "-"
                If MidB$(Content, I + 2, 4) = "->" Then
                    InRemark = False
                    Char = ">"
                End If
                I = I + 4
            Case Else
            End Select
        
        ElseIf InString Then
            If Char = """" Then _
                InString = False _
            Else _
                Buff = Buff & Char
        
        ElseIf InBraket Then
            Select Case Char
            Case "/"
                Select Case LastChar
                Case " "
                Case "<"
                    BracketEnd = True
                    EmptyEnd = False
                    
                    ' trim path
                    If InStrB(Path, "\") Then _
                        Path = Left$(Path, InStrRev(Path, "\") - 1) _
                    Else _
                        Path = ""
                Case Else
                    ' error
                    Exit Function
                End Select
            Case "!"
                If LastChar = "<" Then
                    If MidB$(Content, I + 2, 4) = "--" Then
                        InRemark = True
                        InKey = False
                        InBraket = False
                        Char = "-"
                    Else
                        ' ignore...
                        InKey = False
                    End If
                    I = I + 4
                Else
                    ' error
                    Exit Function
                End If
            Case "?"
                ' definition, ignore...
                InKey = False
            Case ">"
                Select Case LastChar
                Case " "
                Case "/"
                    If BracketEnd Then
                        ' error
                        Exit Function
                    End If
                    
                    BracketEnd = True
                    EmptyEnd = True
                    
                    ' trim path
                    Path = IIf(InStrB(Path, "\"), Left$(Path, InStrRev(Path, "\") - 1), Path)
                Case "?"
                    ' ignore...
                Case Else
                    If Not BracketEnd Then
                        ParseXML = ParseXML( _
                          MidB$(Content, I + 2, LenB(Content) - I - 1), _
                          CurrKey, Path)
                        Exit Function
                    End If
                End Select
                
                InBraket = False
            Case " "
                InKey = False
            Case """"
                Select Case LastChar
                Case " "
                Case "="
                Case Else
                    ' error
                    Exit Function
                End Select
                
                InString = True
                Buff = ""
            Case Else
                If InKey Then _
                    CurrKey = CurrKey & Char
            End Select
        
        Else
            ' not in any structure
            Select Case Char
            Case "<"
                InBraket = True
                InKey = True
                
                If BracketEnd Then
                    BracketEnd = False
                ElseIf LenB(CurrKey) Then
                    Call SetNew(Path & "\" & CurrKey, Trim$(RestoreEntities(CurrValue)))
                    
                    Path = IIf(LenB(Path), Path & "\", "") & CurrKey
                End If
                
                CurrKey = ""
            Case vbCr, vbLf, vbTab, " "
                If LenB(CurrKey) Then
                    If LenB(CurrValue) Then
                        If RightB$(CurrValue, 2) <> " " Then _
                            CurrValue = CurrValue & " "
                    End If
                End If
            Case Else
                If LenB(CurrKey) Then
                    CurrValue = CurrValue & Char
                Else
                    ' error
                    Exit Function
                End If
            End Select
        End If
        
        LastChar = Char
    Next I
    
    ParseXML = True
End Function

Private Function SetNew( _
  Key As String, _
  Value As String _
) As CTreeNode
    Dim Prim    As String, _
        Append  As String
    Dim Slash   As Long
    Dim I       As Long
    
    Slash = InStrB(1, Key, "\")
    If Slash Then _
        Prim = MidB$(Key, Slash + 2, LenB(Key) - Slash - 1)
    
    If ResolveKey(Prim, Append) Then
        Set SetNew = Item(Prim).Add(Append, Value)
    Else
        If Slash = 1 Then
            Set SetNew = Me
            
            mKey = Prim
            mValue = Value
        Else
            Set SetNew = Add(Prim, Value)
        End If
    End If
End Function

'----------------------------------------------------
' the following two functions handle entities that are used in the xml document
' notice that real entities are not recognized!
' this class only can replace numeric entities

Private Function ReplaceEntities(Content As String) As String
    Const SMALLER   As Byte = 60    '<'
    Const GREATER   As Byte = 62    '>'
    Const AMPERSN   As Byte = 38    '&'
    
    Dim I           As Long
    Dim Char        As Byte
    Dim Position    As Long
    Dim Size        As Long
    Dim Entity      As String
    
    Size = LenB(Content)
    
    If Size Then
        ReplaceEntities = String$(Size \ 2, 0)
        Position = 1
        
        For I = 1 To Size Step 2
            Char = AscB(MidB$(Content, I, 2))
            
            Select Case Char
            Case SMALLER, GREATER, Is > 128
                ' resize buffer if necessary
                If Position + 12 > Size Then
                    ReplaceEntities = ReplaceEntities & String$(12, 0)
                    Size = Size + 12
                End If
                
                ' set entity
                Entity = "&#" & CStr(Char) & ";"
            Case AMPERSN
                ' resize buffer if necessary
                If Position + 12 > Size Then
                    ReplaceEntities = ReplaceEntities & String$(12, 0)
                    Size = Size + 12
                End If
                
                ' set entity
                Entity = "&amp;"
            Case Else
                ' all swell
                If Position + 1 > Size Then
                    ReplaceEntities = ReplaceEntities & String$(1, 0)
                    Size = Size + 1
                End If
                
                ' set char
                Entity = Chr$(Char)
            End Select
            
            ' replace char by its entity
            MidB$(ReplaceEntities, Position) = Entity
            Position = Position + LenB(Entity)
        Next I
        
        ReplaceEntities = MidB$(ReplaceEntities, 1, Position - 1)
    End If
End Function

Private Function RestoreEntities(Content As String) As String
    Const SMALLER   As String = "<"
    Const GREATER   As String = ">"
    Const CROSS     As String = "#"
    Const SEMICOLON As String = ";"
    Const AMPERSAND As Byte = 38        '&'
    
    Dim I           As Long
    Dim Char        As Byte
    Dim Position    As Long
    Dim Size        As Long
    Dim Entity      As String
    Dim SemiPos     As Long
    
    Content = Replace(Content, "&amp;", "&")
    
    Size = LenB(Content)
    
    If Size Then
        RestoreEntities = String$(Size \ 2, 0)
        Position = 1
        
        For I = 1 To Size Step 2
            ' do it
            Char = AscB(MidB$(Content, I, 2))
            
            If Char = AMPERSAND Then
                Entity = MidB$(Content, I + 2, 10)
                
                If MidB$(Entity, 1, 2) = CROSS Then
                    SemiPos = InStrB(Entity, SEMICOLON)
                    
                    If SemiPos Then
                        MidB$(RestoreEntities, Position _
                          ) = ChrB$(CByte(MidB$(Entity, 3, SemiPos - 3)))
                        I = I + SemiPos + 1
                    Else
                        MidB$(RestoreEntities, Position) = "&"
                        I = I + 2
                    End If
                    Position = Position + 2
                Else
                    ' already replaced & entity
                    MidB$(RestoreEntities, Position) = ChrB$(Char)
                    Position = Position + 2
                End If
            Else
                ' all swell
                MidB$(RestoreEntities, Position) = ChrB$(Char)
                Position = Position + 2
            End If
        Next I
        
        RestoreEntities = MidB$(RestoreEntities, 1, Position - 1)
    End If
End Function

